@@ -342,6 +342,23 @@ int
   return 1;
 }
 
+static char writeablePath[MAXNAMELEN];
+static char *writeablePathPtr = NULL;
+
+char *EXPORT_CALL
+lbu_setWriteablePath (const char *path)
+{
+  strcpy (writeablePath, path);
+  writeablePathPtr = writeablePath;
+  return writeablePathPtr;
+}
+
+char *EXPORT_CALL
+lbu_getWriteablePath ()
+{
+  return writeablePathPtr;
+}
+
 int EXPORT_CALL
 lbu_charToDots (const char *tableList, const unsigned char *inbuf,
 		unsigned char *outbuf, int length, const char *logFile,
