@@ -89,12 +89,6 @@ alloc_string (const char *inString)
   return newString;
 }
 
-void EXPORT_CALL
-lbu_setWriteablePath (const char *path)
-{
-  ud->writeable_path = alloc_string (path);
-}
-
 int
 file_exists (const char *completePath)
 {
@@ -1053,7 +1047,7 @@ initConfigFiles (const char *firstConfigFile, char *fileName, const
   set_paths (configPath);
   if (logFileName)
     {
-      strcpy ((char *) ud->typeform, ud->writeable_path);
+      strcpy ((char *) ud->typeform, lbu_getWriteablePath ());
       strcat ((char *) ud->typeform, logFileName);
       lou_logFile ((char *) ud->typeform);
     }
@@ -1077,7 +1071,7 @@ read_configuration_file (const char *configFileList, const char
   int listLength;
   int currentListPos = 0;
   errorCount = 0;
-  /*Process logFileName later, after writeable_path is set */
+  /*Process logFileName later, after writeablePath is set */
   if (mode & dontInit)
     {
       ud->has_comp_code = 0;
@@ -1131,7 +1125,9 @@ read_configuration_file (const char *configFileList, const char
 /*Process file list*/
   if (configFileList == NULL)
     {
-      if (!config_compileSettings ("canonical.cfg"))
+      set_paths (NULL);
+      if (!(config_compileSettings ("liblouisutdml.ini") || 
+config_compileSettings ("canonical.cfg")))
 	return 0;
     }
   else
