@@ -71,7 +71,7 @@ static widechar saved_running_head[MAXNAMELEN / 2];
 static widechar saved_footer[MAXNAMELEN / 2];
 static unsigned char saved_typeform[2 * BUFSIZE];
 
-static int
+static void
 saveState (void)
 {
   saved_text_length = ud->text_length;
@@ -109,7 +109,7 @@ saveState (void)
   savePointers ();
 }
 
-static int
+static void
 restoreState (void)
 {
   ud->text_length = saved_text_length;
@@ -161,6 +161,7 @@ transcribe_paragraph (xmlNode * node, int action)
   StyleType *style;
   xmlNode *child;
   int branchCount = 0;
+  int i;
   if (node == NULL)
     return 0;
   if (ud->top == 0)
@@ -216,7 +217,7 @@ transcribe_paragraph (xmlNode * node, int action)
 	pop_sem_stack ();
       return 1;
     case linespacing:
-      do_linespacing;
+      do_linespacing (node);
       if (action != 0)
 	pop_sem_stack ();
       return 1;
@@ -362,7 +363,6 @@ transcribe_paragraph (xmlNode * node, int action)
 	      if (dont_split_this)
 		{
 		  dont_split_status = 1;
-		  int i;
 		  for (i = 1; i < ud->lines_length; i++)
 		    {
 		      if (ud->lines_pagenum[i] > ud->lines_pagenum[i - 1])
@@ -379,7 +379,7 @@ transcribe_paragraph (xmlNode * node, int action)
 		    orphan_control_status = 1;
 		  else
 		    orphan_control_status = 0;
-		  int i = 1;
+		  i = 1;
 		  while (i < orphan_control
 			 && orphan_control_pos + i < ud->lines_length)
 		    {
